"use strict";
// This file is part of classy-mst, copyright (c) 2017- BusFaster Ltd.
// Released under the MIT license, see LICENSE.
Object.defineProperty(exports, "__esModule", { value: true });
var mobx_state_tree_1 = require("mobx-state-tree");
exports.typeTag = '$';
function setTypeTag(tag) {
    exports.typeTag = tag;
}
exports.setTypeTag = setTypeTag;
function dummyGetter() { }
/** Force TypeScript to accept an MST model as a superclass.
  * @param model Model (MST tree node) */
function shim(Model, Parent) {
    function BaseClass() { }
    if (Parent && Parent.$proto) {
        BaseClass.prototype = Parent.$proto;
        BaseClass.prototype = new BaseClass();
        BaseClass.prototype.$parent = Parent;
        if (BaseClass.prototype.$actions)
            BaseClass.prototype.$actions = {};
    }
    else {
        BaseClass.prototype = {};
    }
    return BaseClass;
}
exports.shim = shim;
/** Decorator for actions. By default the mst function treats methods as views. */
function action(target, key) {
    (target.$actions || (target.$actions = {}))[key] = true;
}
exports.action = action;
var internalMembers = {
    constructor: true,
    $actions: true,
    $parent: true
};
function renameFunction(func, name) {
    Object.defineProperty(func, 'name', { value: name, writable: false });
}
// Test if renaming functions works to avoid errors / useless attempts.
try {
    renameFunction(dummyGetter, 'dummy');
}
catch (err) { }
var renamableFunctions = (dummyGetter.name == 'dummy');
/** Add methods from an ES6 class into an MST model.
  * @param code Class with methods to add as views (the default)
  *   and actions (if decorated).
  * @param data MST model with properties.
  * @param modelName Model name for debugging and polymorphic type tags in snapshots. */
function mst(Code, Data, modelName, options) {
    var viewList = [];
    var actionList = [];
    var descList = [];
    var memberTbl = Code.prototype;
    var actionTbl = memberTbl.$actions;
    var volatileTbl = {};
    if (modelName && typeof (modelName) == 'object') {
        options = modelName;
        modelName = options.name;
    }
    else
        options = options || {};
    // Extract views, actions, getters and setters from the class prototype.
    for (var _i = 0, _a = Object.getOwnPropertyNames(memberTbl); _i < _a.length; _i++) {
        var name = _a[_i];
        if (internalMembers[name])
            continue;
        var desc = Object.getOwnPropertyDescriptor && Object.getOwnPropertyDescriptor(memberTbl, name);
        if (!desc || (desc.configurable && desc.writable && !desc.get && !desc.set)) {
            var value = memberTbl[name];
            var spec = { name: name, value: value };
            if (actionTbl && actionTbl[name])
                actionList.push(spec);
            else
                viewList.push(spec);
        }
        else {
            descList.push({ name: name, value: desc });
        }
    }
    // Create a sample instance and extract volatile members
    // defined in the constructor.
    var instance = new Code();
    var volatileList = Object.getOwnPropertyNames(instance);
    for (var _b = 0, volatileList_1 = volatileList; _b < volatileList_1.length; _b++) {
        var name = volatileList_1[_b];
        volatileTbl[name] = instance[name];
    }
    // Apply optional name given to the model.
    if (modelName)
        Data = Data.named(modelName);
    // Bind views, actions and volatile state to the model.
    var Model = Data;
    Model = options.sealed ? Model : Model.preProcessSnapshot(
    // Instantiating a union of models requires a snapshot.
    function (snap) { return snap || {}; });
    Model = Model.postProcessSnapshot(function (snap) {
        if (modelName && exports.typeTag && Code.prototype.$parent)
            snap[exports.typeTag] = modelName;
        return (snap);
    }).actions(function (self) {
        var result = {};
        var _loop_1 = function (name, value) {
            var method = function () {
                return (value.apply(self, arguments));
            };
            if (renamableFunctions) {
                renameFunction(method, modelName + '.' + name);
            }
            result[name] = method;
        };
        for (var _i = 0, actionList_1 = actionList; _i < actionList_1.length; _i++) {
            var _a = actionList_1[_i], name = _a.name, value = _a.value;
            _loop_1(name, value);
        }
        return (result);
    });
    Model = !(viewList.length + descList.length) ? Model : Model.views(function (self) {
        var result = {};
        var _loop_2 = function (name, value) {
            result[name] = function () {
                return (value.apply(self, arguments));
            };
        };
        for (var _i = 0, viewList_1 = viewList; _i < viewList_1.length; _i++) {
            var _a = viewList_1[_i], name = _a.name, value = _a.value;
            _loop_2(name, value);
        }
        var _loop_3 = function (name, value) {
            var desc = {};
            for (var _i = 0, _a = Object.getOwnPropertyNames(value); _i < _a.length; _i++) {
                var key = _a[_i];
                desc[key] = value[key];
            }
            var get = desc.get, set = desc.set;
            if (get) {
                desc.get = function () { return get.call(self); };
            }
            else if (set) {
                // Properties with only setters still need a getter defined here,
                // or mobx-state-tree will ignore them.
                desc.get = dummyGetter;
            }
            if (set)
                desc.set = function (value) { return set.call(self, value); };
            // Allow mobx-state-tree to see the property, or it gets ignored.
            desc.enumerable = true;
            Object.defineProperty(result, name, desc);
        };
        for (var _b = 0, descList_1 = descList; _b < descList_1.length; _b++) {
            var _c = descList_1[_b], name = _c.name, value = _c.value;
            _loop_3(name, value);
        }
        return (result);
    });
    Model = !volatileList.length ? Model : Model.volatile(function (self) { return volatileTbl; });
    return (options.sealed ? Model : polymorphic(Code, Model, modelName));
}
exports.mst = mst;
function polymorphic(Code, Model, modelName) {
    // Union of this class and all of its subclasses.
    // Late evaluation allows subclasses to add themselves to the type list
    // before any instances are created.
    var Union = mobx_state_tree_1.types.late(function () { return mobx_state_tree_1.types.union.apply(mobx_state_tree_1.types, Union.$typeList); });
    // First item in the type list is a dispatcher function
    // for parsing type tags in snapshots.
    Union.$typeList = [{ dispatcher: function (snap) {
                return (snap && exports.typeTag && snap[exports.typeTag] && Union.$typeTbl[snap[exports.typeTag]]) || Model;
            }
        }];
    Union.$typeTbl = {};
    Union.$proto = Code.prototype;
    // Copy methods from model object into returned union,
    // making it work like a regular model.
    for (var mixin = Model; mixin; mixin = Object.getPrototypeOf(mixin)) {
        var _loop_4 = function (key) {
            var desc = Object.getOwnPropertyDescriptor && Object.getOwnPropertyDescriptor(mixin, key);
            if (!desc || (desc.configurable && desc.writable && !desc.get && !desc.set)) {
                var value_1 = !(key in Union) && mixin[key];
                if (typeof (value_1) == 'function') {
                    Union[key] = function () {
                        return (value_1.apply(Model, arguments));
                    };
                }
            }
        };
        for (var _i = 0, _a = Object.getOwnPropertyNames(mixin); _i < _a.length; _i++) {
            var key = _a[_i];
            _loop_4(key);
        }
    }
    // Initialize union of allowed class substitutes with the class itself,
    // and augment unions of all parent classes with this subclass,
    // to allow polymorphism.
    for (var Class = Union; Class; Class = Class.$proto.$parent) {
        var typeList = Class.$typeList;
        var typeTbl = Class.$typeTbl;
        if (typeList)
            typeList.push(Model);
        if (typeTbl && modelName)
            typeTbl[modelName] = Model;
    }
    return (Union);
}
exports.polymorphic = polymorphic;
function mstWithChildren(Code, Data, name) {
    var Children = mobx_state_tree_1.types.array(mobx_state_tree_1.types.late(function () { return Model; }));
    var Branch = Data.props({
        children: mobx_state_tree_1.types.maybe(Children)
    });
    var Model = mst(Code, Branch, 'Node');
    return ({ Model: Model, Children: Children });
}
exports.mstWithChildren = mstWithChildren;
